package SM.diglib;

import org.w3c.rdf.*;
import org.ginf.*;
import org.ginf.helpers.*;
import org.ginf.delivery.*;
import org.ginf.delivery.HTTP.*;
import org.ginf.schemas.*;
import org.xml.sax.*;

import java.io.*;
import java.net.*;
import java.util.*;

public class Server implements Layer {

  RDFFactory factory;
  Layer outside;
  Model db;

  public Server(RDFFactory factory, URL url) throws Exception {
    this.factory = factory;
    SchemaRegistry reg = factory.registry();

    reg.addLoader( new DublinCore() ); // do not attempt to download DublinCore schema
    reg.addLoader( new Coll() );
    reg.addLoader( new SDLIP() );

    // load database
    RDFMS p = factory.createRDFMS();
    InputSource source = new InputSource( url.openStream() );
    source.setSystemId( url.toString() );
    db = p.parse( source );
  }

  Resource getStateInfoClass() {
    return HTTP.Connection;
  }

  public void connect(String what, Layer layer) {
    if("transport".equals(what))
      this.outside = layer;
    else
      throw new IllegalArgumentException("[Client] cannot connect \"" + what + "\"");
  }

  public String toString() {
    return "SDLIP Server Layer";
  }

  /**
   * Layer implementation
   */
  public Model dispatch(Model model, Layer from) throws Exception {

    if(from != outside)
      return null; // we can only process client requests

    //    System.err.println("GOT CLIENT REQUEST");

    // we want to access schema info
    SchemaModel m = factory.createSchemaModel( model );

    //    System.err.println("\nGOT MODEL: " + SDLIP.DiscardState);
    //    RDFUtil.printTriples( m, System.err );

    if( m.find( null, RDF.type, CoreState.DiscardState ).size() > 0 ) {
      System.err.println("[DLServer] DISCARDING STATE");
      return null;
    }

    // find search request
    Resource s = m.find1( null, RDF.type, CoreComm.Message ).subject();
    Resource state = m.find1( null, RDF.type, getStateInfoClass() ).subject();

    if(!RDFUtil.instanceOf(s, SDLIP.Search, m)) {
      error(m, state, SDLIP.Error, "Unknown request");
      return null;
    }
    // validate the model
    //    System.err.println("\nGOT MODEL");
    //    RDFUtil.printTriples( m, System.err );
    //    System.err.println("\nREGISTRY (DOMAINS)");
    //    RDFUtil.printTriples( factory.registry().get().find( null, SchemaRegistry.domain, null), System.err );

    // factory.createRDFMS().serialize( m, new OutputStreamWriter(System.out) );
    m.validate();


    String query = null;
    int limit = 0;
    Vector neededAttrs = new Vector();

    try {
      //      Resource s = t.subject(); // search request


      System.out.println("[DLServer] REQUEST FROM: " +
			 m.find1( null, CoreComm.from, null ).object());

      // query
      Triple qt = m.find1( s, SDLIP.query, null );
      if(qt != null)
	query = qt.object().toString().toUpperCase();

      // limit
      qt = m.find1( s, SDLIP.limit, null );
      if(qt != null)
	try {
	  limit = Integer.parseInt( qt.object().toString() );
	} catch (Exception pe) {}

      System.out.println("The query is: " + query );
      System.out.println("Limit is: " + limit );

      // find Bag with requested attributes
      Resource attrs = (Resource)m.find1( s, SDLIP.attributes, null ).object();
      //      System.out.println("Requested attributes:");
      neededAttrs = RDFUtil.getContainerProperties( attrs, m );

    } catch (Exception any) {}

    // GENERATE REPLY
    // this will be the reply
    Model reply = factory.createModel();
    Resource result = createReply( m, reply, state, SDLIP.Result );

    // generate Warnings
    for(int i=0; i < neededAttrs.size(); i++) {
      Resource att = (Resource)neededAttrs.elementAt(i);
      if(!(DublinCore.Title.equals(att) || DublinCore.Creator.equals(att) || DublinCore.Date.equals(att))) {
	reply.add( result, Coll.unsupportedAttribute, att );
      }
    }

    // find all documents
    Model list = db.find( null, DublinCore.Title, null );

    int count = 0;
    for(Enumeration en = list.elements(); en.hasMoreElements(); ) {

      if(limit > 0 && count >= limit)
	break;

      Resource doc = ((Triple)en.nextElement()).subject();
      // matches
      Triple key = db.find1( doc, DublinCore.Source, null );
      String keywords = null;
      if(key != null)
	keywords = key.object().toString();
      else {
	try {
	  keywords = (db.find1( doc, DublinCore.Title, null).object().toString() + " " +
		      db.find1( doc, DublinCore.Creator, null).object()).toUpperCase();
	  db.add( doc, DublinCore.Source, keywords );
	} catch (Exception any) {}
      }

      if(query != null && keywords.indexOf(query) >= 0) {
	// append to the result set
	count++;
	reply.add( doc, RDF.type, SDLIP.ResultItem );
	reply.add( doc, SDLIP.rank, String.valueOf(count) );
	reply.add( result, SDLIP.resultItem, doc ); // append to the result set
	// take attributes
	if(neededAttrs.size() == 0) {
	  // get all of them
	  reply.union( db.find( doc, DublinCore.Title, null ) );
	  reply.union( db.find( doc, DublinCore.Creator, null ) );
	} else
	  for(int j=0; j < neededAttrs.size(); j++) {
	    Resource prop = (Resource)neededAttrs.elementAt(j);
	    Triple att = db.find1( doc, prop, null );
	    if(att != null)
	      reply.add( att ); // doc, prop, att.object() );
	  }
      }
    }

    reply.add( result, SDLIP.resultSize, String.valueOf(count) ); // only two as an example
    
    from.dispatch( reply, this );

    // hang up
    reply = factory.createModel();
    createReply(m, reply, state, CoreState.DiscardState);
    from.dispatch(reply, this);

    return null;
    // m.merge(envelope);
    // comm.send(m);
  }
  /*
  void error(String text, Model orig, Layer from) {

    try {
      Model m = factory.createModel();
      Resource e = createReply( orig, m, SDLIP.Error );
      m.add( e, CoreComm.errorMessage, text );
      from.dispatch( m, this );
    } catch (Exception any) {}
  }

  public Resource createReply(Model orig, Model reply, Resource replyType) {
    
    Resource r = orig.find1( null, RDF.type, CoreComm.Request ).subject();
    Resource e = RDFUtil.noname();
    reply.add( e, RDF.type, replyType );
    reply.add( e, CoreComm.inReplyTo, r );
    reply.add( r, RDF.type, CoreComm.Request );
    return e;
  }
  */

  void error(Model orig, Resource state, Resource errorType, String text) {
    
    try {
      Model m = factory.createModel();
      Resource e = createReply(orig, m, state, errorType);
      m.add( e, CoreComm.errorMessage, text );
      outside.dispatch(m, this);
      // hang up
      m = factory.createModel();
      createReply(orig, m, state, CoreState.DiscardState);
      outside.dispatch(m, this);
    } catch (Exception any) {}
  }

  public Resource createReply(Model orig, Model reply, Resource state, Resource replyType) {
    
    Resource r = orig.find1( null, RDF.type, CoreComm.Request ).subject();
    Resource e = RDFUtil.noname();
    reply.add( e, RDF.type, replyType );
    reply.add( e, CoreState.state, state );
    reply.add( state, RDF.type, CoreState.StateInfo );
    return e;
  }

  public static void main(String[] args) {

    if (args.length == 0) {
      System.err.println("Usage: java -Dorg.xml.sax.parser=<classname> SM.diglib.Server [ URI | filename ]");
      System.exit(1);
    }

    try {

      URL url = null;
      try {
	url = new URL (args[0]);
      } catch (Exception e) {
	url = new URL ("file", null, args[0]);
      }

      RDFFactory factory = new RDFFactoryImpl();

      Server s = new Server(factory, url);
      HTTPLayer h = new HTTPLayer( factory );
      //      ConnectionlessTransport t = new ConnectionlessTransport( factory, true );
      
      h.connect( "dispatch", s );
      //      t.connect( "http", h );
      //      t.connect( "service", s );
      s.connect( "transport", h );
      
      h.acceptConnections();

    } catch (Exception any) {
      System.err.println("[DLServer] Could not launch server: " + any);
      any.printStackTrace(System.err);
    }
  }
  
}
