package SM.ginf;

import org.w3c.rdf.*;
import org.ginf.*;
import org.ginf.helpers.*;
import org.ginf.delivery.HTTP.*;
import org.ginf.delivery.*;
import org.ginf.schemas.*;

import java.util.*;
import java.io.*;

public class HelloWorldClient2 implements Layer {

  RDFFactory factory;
  Layer transport;

  public HelloWorldClient2(RDFFactory factory) throws Exception {

    this.factory = factory;
    factory.registry().addLoader(new HelloWorld());
  }

  public void connect(String what, Layer layer) {
    if("transport".equals(what))
      this.transport = layer;
    else
      throw new IllegalArgumentException(toString() + " cannot connect \"" + what + "\"");
  }

  /**
   * Layer implementation
   */

  public Model dispatch(Model model, Layer from) throws Exception {

    // we want to access schema information
    SchemaModel m = factory.createSchemaModel( model );

    try {
      // Optionally validate schema conformance
      m.validate();

      // find the reply
      Resource reply = m.find1( null, RDF.type, CoreComm.Message ).subject();
      
      // we can process only StringToPrint repplies
      if(m.find( reply, RDF.type, HelloWorld.StringToPrint ).size() > 0) {
	
	System.out.println( m.find1( reply, HelloWorld.string, null ).object().toString() );
      }
    } catch (Exception any) {
      System.err.println(toString() + " ERROR " + any.getMessage());
    }
    // done
    return null;
  }

  void sendRequest(String url) throws Exception {

    // we construct the model using the API, so fetch the schema manually
    factory.registry().fetch(HelloWorld.NS);

    Model m = factory.createModel();
    
    // create new unique unnamed resource
    Resource s = RDFUtil.noname();
    // this is our request
    m.add( s, RDF.type, HelloWorld.GetString );
    // destination
    m.add( s, CoreComm.to, new Resource(url));
    m.add( new Resource(url), RDF.type, CoreComm.ProcessingEntity );

    transport.dispatch(m, this);
  }

  public String toString() {
    return "Hello World Client";
  }

  public static void main(String[] args) throws Exception {

    // create the default factory
    RDFFactory factory = new RDFFactoryImpl();
    
    // create the client
    HelloWorldClient2 client = new HelloWorldClient2(factory);
    // create the transport layer
    HTTPLayer h = new HTTPLayer(factory);
    client.connect( "transport", h );
    
    client.sendRequest("http://localhost");
  }
}
