/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package org.ginf;

import org.xml.sax.InputSource;
import org.w3c.rdf.Resource;
import java.io.Writer;

/**
 * This is an interface of an RDF parser/serializer.
 *
 * @author Sergey Melnik <melnik@db.stanford.edu>
 */

public interface RDFMS {

  /**
   * Parse from the given SAX/XML input source.
   * Use factory to create models and access the schema registry.
   *
   * @see    org.ginf.RDFFactory
   */
  public Model parse(InputSource source) throws Exception;

  /**
   * Overloads standard schema loading policy of the parser
   * (which could be, for example, to call schema registry to fetch unknown
   * schemas).
   */
  public void setSchemaController(SchemaController c);

  /**
   * Serialize a given model into an XML character output stream.
   */
  public void serialize(Model model, Writer w) throws Exception;

}
