/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */
package org.ginf;

/**
 * An interface to a schema-aware model. Such a model
 * extends the basic model interface by providing an additional
 * validation method. <code>find</code> method must be
 * implemented using subclass and subproperty information contained
 * in the schema. Thus, a schema-aware model must compute
 * the transitive closure on the predicates <code>subClassOf</code>
 * and <code>subPropertyOf</code>.
 * 
 * <p>Given the triples:<br>
 * <pre><blockquote>
 * (r1,     RDF.type,           ClassA)
 * (ClassA, RDFS.subClassOf,    ClassB)
 * (ClassB, RDFS.subClassOf,    ClassC)
 * (r1,     propA,              r2)
 * (propA,  RDFS.subPropertyOf, propB)
 * (propB,  RDFS.subPropertyOf, propC)
 * </blockquote></pre>
 * we must be able to derive:
 * <pre><blockquote>
 * (r1,     RDF.type,         ClassB)
 * (r1,     RDF.type,         ClassD)
 * (ClassA, RDFS.subClassOf,    ClassD)
 * (propA,  RDFS.subPropertyOf, propC)
 * (r1,     propB,              r2)
 * (r1,     propC,              r2)
 * </blockquote></pre>
 *
 * @author Sergey Melnik <melnik@db.stanford.edu>
 */


public interface SchemaModel extends Model {

  /**
   * Validates the model according to the schema specification.
   */
  public void validate() throws Exception;
}
