/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package org.ginf.delivery.HTTP;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.ginf.*;
import org.ginf.helpers.*;
import org.ginf.schemas.CoreComm;
import org.ginf.schemas.HTTP;
import org.ginf.schemas.CoreState;
import org.ginf.schemas.RDF;
import org.ginf.delivery.*;
import org.w3c.rdf.*;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * A new Connection object is created for each new connection
 * established between the server and a client.
 *
 * @author Sergey Melnik <melnik@db.stanford.edu>
 */

public class Connection extends Thread {
  
  HTTPLayer       http;
  Socket          socket;
  InputStream     in;
  OutputStream    out;
  
  Layer inside;

  Resource state;
  //  String to; // once forward smth store this "to" info and do not override it any more

  boolean initiatedFromOutside = false;
  // only makes sense for the server side
  int replyCount = 0;

  // only makes sense for the client side
  String BOUNDARY;
  
  boolean firstReply = true; // write HTTP / OK or not

  boolean notify = true; // notify layer if IO error
  
  // The communication streams are initiated and the thread is started.
  Connection(HTTPLayer http, Layer inside, Resource state,
	     Socket socket, ThreadGroup threadgroup, int priority) {

    // The Connection thread is given a group, a name, and a priority.
    super(threadgroup, "HTTP Connection");
    
    //      System.out.println("New connection from " + client_socket);
    
    this.http = http;
    this.setPriority(priority);
    this.socket = socket;
    this.inside = inside;
    this.state = state;

    // The data-exchange streams are created.
    try {
      in = socket.getInputStream();
      out = socket.getOutputStream();
      //      in = new BufferedInputStream(socket.getInputStream());
      //      out = new BufferedOutputStream(socket.getOutputStream());
    } catch (IOException e) {
      try {
	socket.close();
      } catch (IOException e2) { ; }
	
      System.err.println(http + " Exception while getting socket streams: " + e);
      return;
    }
    // The Connection thread is started.
    this.start();
  }

  //  String locator;
  /*
  void getLocator(String line) {

    int i1 = line.indexOf(' ');
    int i2 = line.indexOf(' ', i1+1);
    locator = line.substring(i1+1, i2);
  }
  */

  // reads header, field names are in lower case
  Hashtable getMIMEHeader(BufferedReader din) throws IOException {

    Hashtable header = new Hashtable();

    do {
      String line = din.readLine();
      if(line == null) {
	//	http.sendError("No HTTP header", this);
	break; // error
      }
      if(line.length() == 0)
	break; // header ended
      // find colon
      int colon = line.indexOf(':');
      if(colon < 1)
	break; // error
      header.put( line.substring(0, colon).toLowerCase(), line.substring(colon + 2) );
      //	  System.err.println("HEADER: " + line.substring(0, colon) + "=" + line.substring(colon+2));
    } while (true);
    
    return header;
  }

  String getBoundary(String value) {
    
    if(value == null)
      return null;
    int i = value.indexOf("boundary=");
    if(i < 0)
      return null;
    else
      return value.substring(i + 9);
  }

  public void run() {
    
    try {
      
      // wait until client is done
      loop:while(true) {
	
	// read header
	//	Hashtable header = new Hashtable(); // type->value

	//	SM.util.FileUtils.copyStream( in, System.err, 1024 );

	boolean encoded = false, hasParams = false; // valid for subsequent replies

	Model model = null;
	InputSource source = null;
	BufferedReader din = new BufferedReader( new InputStreamReader(in) );
	// DataInputStream din = new DataInputStream( in );
	  
	if(initiatedFromOutside) {

	  // SM.util.FileUtils.copyCharStream( din, new OutputStreamWriter(System.err), 1024 );

	  String line = din.readLine();
	  // System.err.println("--- LINE: " + line);

	  if(line.startsWith("POST")) {

	    //	  System.err.println("VALID HTTP request: " + line);
	    // ignore header
	    getMIMEHeader(din);
	    
	    // standard parameter: RDF=
	    // read until the first parameter
	    while( (char)din.read() != '=') {}
	    source = new InputSource( new HTTPDecoder( din ) );
	    
	  } else if(line.startsWith("GET")) {
	    // do nothing, source remains null
	  } else {
	    //	    System.err.println("INVALID HTTP request: " + line);
	    http.sendError("Invalid HTTP request: \"" + line + "\"", this);
	    break loop; // error
	  }
	  
	} else {
	  // initiated locally

	  // prepare reader

	  // SM.util.FileUtils.copyCharStream( din, new OutputStreamWriter(System.err) );
	  //	  SM.util.FileUtils.copyStream( in, System.err );

	  ReadUntilFilter ruf = new ReadUntilFilter(din);

	  if(replyCount++ == 0) {
	    
	    // this is the first reply we get
	    String line = din.readLine();

	    // System.err.println("---line: " + line);

	    if(!line.startsWith("HTTP")) {
	      http.sendError("Not an HTTP reply header: " + line, this);
	      break loop;
	    }
	    Hashtable header  = getMIMEHeader(din);
	    BOUNDARY = getBoundary( (String)header.get("content-type") );
	    
	    if(BOUNDARY == null) {
	      http.sendError("No valid multipart header found", this);
	      break loop;
	    }
	    // System.err.println("---BOUNDARY: " + BOUNDARY);

	    ruf.initDelimiter("--" + BOUNDARY);


	    // skip first boundary
	    int c;
	    while ((c = ruf.read()) >= 0) { /*System.err.print("" + (char)c);*/ }
	    din.readLine(); // boundary is followed by a line break
	    ruf.initDelimiter();

	    // SM.util.FileUtils.dumpCharStream( din, 150 );
	    // SM.util.FileUtils.copyCharStream( ruf, new OutputStreamWriter(System.err) );


	  } else // just init
	    ruf.initDelimiter("--" + BOUNDARY);

	  getMIMEHeader(din);
	  // now we are ready to parse RDF/XML
	  source = new InputSource( ruf );
	}

	//	SM.util.FileUtils.copyStream( source.getByteStream(), System.err, 1024 );
	//	System.err.print("Written out request:");
	
	if(source != null) { // null if GET
	  
	  RDFMS p = http.factory.createRDFMS();
	  if(http.controller != null) // override default
	    p.setSchemaController(http.controller);
	  
	  model = p.parse( source );
	  //	  System.err.println(http + " GOT MODEL " + model.size());

	} else {
	  System.err.println(http + " EMPTY MODEL");
	  // create empty model
	  model = http.factory.createModel();
	}
	
	model = http.factory.createSchemaModel( model );

	Resource message = RDFUtil.get1(model.find( null, RDF.type, CoreComm.Message )).subject();
	// FIXME: if multiple AND different Messages -> error

	if(message == null) {
	  // no message -> create one!
	  message = RDFUtil.noname();
	  model.add( message, RDF.type, CoreComm.Request ); // FIXME: just a Message or Request?
	}

	//	System.err.println("---MY STATE IS: " + state);
	// determine the state or create a new state
	if(state == null) {
	  // can we borrow the state from the other side of the connection?
	  Triple t_state = RDFUtil.get1(model.find( null, RDF.type, HTTP.Connection ));
	  if(t_state != null)
	    state = t_state.subject();
	  else {
	    // none, create my own
	    state = RDFUtil.noname();
	    model.add( state, RDF.type, HTTP.Connection );
	    model.add( message, CoreState.state, state );
	    //	      }
	  }
	  // register state
	  http.connPool.put( state, this );
	  //	  System.err.println("---REGISTERING FOREIGN STATE (FOREIGN INITIATED): " + state);

	} else {
	  // we have our own state. If it is not in the reply, append it
	  // and remove any other HTTP.Connection inside
	  Triple t_state = RDFUtil.get1(model.find( null, RDF.type, HTTP.Connection ));

	  /*
	  if(t_state != null)
	    System.err.println("---GOT ALTERNATIVE STATE: " + t_state.subject() + ", MINE: " + state);
	  else
	    System.err.println("---NO ALTERNATIVE STATE, TAKING MINE: " + state);
	    */

	  if(t_state != null && !state.equals(t_state.subject())) {
	    model.remove(t_state);
	    model.remove( model.find1( null, CoreState.state, t_state.subject() ) );
	  }
	  if(t_state == null || !state.equals(t_state.subject())) {
	    model.add( message, CoreState.state, state );
	    model.add( state, RDF.type, HTTP.Connection );
	  }
	}

	//	System.err.println("CALLING DISPATCHER...");
	Resource pe = new Resource("http://" + socket.getInetAddress().getHostName() + ":" + socket.getPort());
	model.add( message, HTTP.from, pe );
	model.add( pe, RDF.type, CoreComm.ProcessingEntity );
	
	inside.dispatch( model, http );
      }
      
    } catch (IOException eof) {
      // terminate connection
      http.sendError( "Invalid communication attempt " + eof, this );

    } catch (SAXException any) {
      http.sendError( "Invalid RDF/XML", this );

    } catch (Exception any) {

      if(notify) {
	System.err.println(http + " Dispatch error: " + any.getClass() + ", " + any);
	any.printStackTrace(System.err);
      }
      
    } finally {
      try {
	http.connPool.remove( this );

	// notify of discarding
	if(notify) {
	  Model reply = http.factory.createModel();
	  Resource s = RDFUtil.noname();
	  reply.add( s, RDF.type, CoreState.DiscardState );
	  reply.add( s, CoreState.state, state );
	  //	  if(to != null)
	  //	    reply.add( s, CoreComm.to, to );
	  reply.add( state, RDF.type, HTTP.Connection );
	  // FIXME: to
	  inside.dispatch( reply, http );
	  //	  getDispatcher( locator ).dispatch( reply, http );
	}

      } catch (Exception e2) { ; }
      try {
	//	System.err.println(http + " Hanging up connection.");
	socket.close();
      } catch (IOException e3) {}
    }
  }

  public void close() {

    notify = false;
    try {
      in.close();
      out.close();
      socket.close();
    } catch (IOException any) {}
  }

}

